/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

unsigned long slb_loop_serial;
time_t slb_loop_ts;
struct slb_server **srvtab;
size_t srvtab_count;

static size_t active_count;

char *
srvid(struct slb_server *srv)
{
	return srv->id ? srv->id : srv->peer;
}

#define E_VARINST_OK 0
#define E_VARINST_UNSUPPORTED 1
#define E_VARINST_OVERFLOW 2

static int
varinst_assign(struct slb_varinstance *inst, struct variable_list *var,
	       struct slb_server *srv)
{
	int rc = E_VARINST_OK;
	double val;
	
	switch (var->type) {
	case ASN_INTEGER:
		inst->vi_value = *var->val.integer;
		break;
			
	case ASN_UINTEGER:
		inst->vi_value = (unsigned) *var->val.integer;
		break;

	case ASN_GAUGE:
	case ASN_COUNTER:
		val = (unsigned) *var->val.integer;
		if (val < inst->vi_value)
			rc = E_VARINST_OVERFLOW;
		else
			inst->vi_value = val;
		break;
		
	case ASN_COUNTER64:
		val = (double) var->val.counter64->high * 4294967296.0 +
			             var->val.counter64->low;
		if (val < inst->vi_value)
			rc = E_VARINST_OVERFLOW;
		else
			inst->vi_value = val;
		break;
		
#ifdef NETSNMP_WITH_OPAQUE_SPECIAL_TYPES
	case ASN_OPAQUE_FLOAT:
		inst->vi_value = *var->val.floatVal;
		break;
		
	case ASN_OPAQUE_DOUBLE:
		inst->vi_value = *var->val.doubleVal;
		break;
#endif
		
	default:
		logmsg(LOG_ERR,
		       "server %s: variable %s has unsupported type (%d)",
		       srvid(srv), inst->vi_sym.name,
		       var->type);
		return E_VARINST_UNSUPPORTED;
	}

	if (rc == E_VARINST_OVERFLOW)
		logmsg(LOG_NOTICE,
		       "server %s: variable %s overflow (old=%f, new=%f)",
		       srvid(srv), inst->vi_sym.name,
		       inst->vi_value, val);
	
	return rc;
}

static int
assertion_test(struct slb_assertion *ap, struct variable_list *vp)
{
	size_t out_len = 0;
	int rc;

	if (ap->bufsize == 0) {
		/* Initial allocation */
		ap->bufsize = 16;
		ap->buf = grecs_malloc(ap->bufsize);
	}
	if (!sprint_realloc_value(&ap->buf, &ap->bufsize, &out_len, 1, ap->name,
				  ap->length, vp))
		return 1;
	rc = strcmp(ap->value, (char*) ap->buf);
	if (ap->flags & SLB_ASSERT_NE)
		rc = !rc;
	return rc;
}

static void
process_result(struct slb_server *srv, struct snmp_pdu *pdu)
{
	struct variable_list *vp;
	
	debug(SLB_DEBCAT_SNMP, 2, ("reply from %s", srvid(srv)));
	for (vp = pdu->variables; vp; vp = vp->next_variable) {
		struct slb_assertion *ap;
		struct slb_varinstance *inst;
		char buf[1024];

		ap = assertion_lookup(srv->assertions, vp->name,
				      vp->name_length, NULL);
		if (ap && assertion_test(ap, vp)) {
			logmsg(LOG_NOTICE,
			       "%s:%d: %s: assertion \"%s\" failed",
			       ap->locus.file, ap->locus.line, srvid(srv),
			       ap->text);
			return;
		}
		
		inst = oidtab_lookup(srv->oidtab, vp->name, vp->name_length);
		if (!inst) {
			if (ap)
				continue;
			snprint_variable(buf, sizeof(buf),
					 vp->name, vp->name_length, vp);
			logmsg(LOG_CRIT,
			       "server %s: could not find variable %s!",
			       srvid(srv), buf);
			abort();
		}

		switch (varinst_assign(inst, vp, srv)) {
		case E_VARINST_OK:
			inst->vi_flags |= SLV_VIF_SET;
			if (debug_level[SLB_DEBCAT_SNMP] >= 3) {
				snprint_variable(buf, sizeof(buf),
						 vp->name, vp->name_length,
						 vp);
				debug_printf("%s: %s = %s",
					     srvid(srv), inst->vi_sym.name,
					     buf);
			}
			break;

		case E_VARINST_UNSUPPORTED:
			break;

		case E_VARINST_OVERFLOW:
			slb_node_state_init(srv->state);
			srv->expr->ex_eval_count = 0;
			varinst_clear(srv->varinst);
			return;
		}
	}

	if (varinst_allset(srv->varinst)) {
		debug(SLB_DEBCAT_SNMP, 3, ("%s: evaluating expression: \"%s\"",
			  srvid(srv), srv->expr->ex_text));
		if (eval_expression(srv->expr, &srv->state, &srv->weight)
		    == 0) {
			srv->flags |= SLB_SRV_COMPUTED;
			debug(SLB_DEBCAT_SNMP, 3, ("%s: expression yields %g",
				  srvid(srv), srv->weight));
		}
	}
}

void
server_process_pdu(int operation, struct slb_server *srv, struct snmp_pdu *pdu)
{
	if (operation == NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE) {
		process_result(srv, pdu);
	} else {
		logmsg(LOG_ERR, "server %s: timeout", srvid(srv));
		/* Re-initialize server state */
		slb_node_state_init(srv->state);
		srv->expr->ex_eval_count = 0;
	}
}

int
asynch_response(int operation, struct snmp_session *sp, int reqid,
		struct snmp_pdu *pdu, void *magic)
{
	struct slb_server *srv = magic;

	server_process_pdu(operation, srv, pdu);
	active_count--;
	return 1;
}

static int
add_snmp_var(void *sym, void *data)
{
	struct slb_varinstance *vinst = sym;
	struct snmp_pdu *req = data;

	if ((vinst->vi_flags & (SLV_VIF_SNMP|SLV_VIF_USED)) ==
	    (SLV_VIF_SNMP|SLV_VIF_USED))
		snmp_add_null_var(req, vinst->vi_oid, vinst->vi_oidlen);
	return 0;
}

static int
add_snmp_assertion(void *sym, void *data)
{
	struct slb_assertion *ap = sym;
	struct snmp_pdu *req = data;

	snmp_add_null_var(req, ap->name, ap->length);
	return 0;
}

static void
send_requests()
{
	struct slb_server *srv;

	debug(SLB_DEBCAT_SNMP, 1, ("sending requests"));
	for (srv = srv_head; srv; srv = srv->next) {
		struct snmp_session sess;
		struct snmp_pdu *req;

		if (srv->flags & SLB_SRV_DISABLED)
			continue;

		debug(SLB_DEBCAT_SNMP, 2,
		      ("server %s, host %s", srvid(srv), srv->peer));
		snmp_sess_init(&sess);
		sess.version = srv->snmpver;
		sess.peername = srv->peer;
		if (srv->community) {
			sess.community = srv->community;
			sess.community_len = strlen((char*)sess.community);
		}
		if (srv->timeout)
			sess.timeout = srv->timeout * 1000000L;
		if (srv->retries)
			sess.retries = srv->retries;
		if (sess.version == SNMP_VERSION_3) {
			sess.securityAuthProto =
				srv->snmp_v3.auth.proto;
			sess.securityAuthProtoLen =
				srv->snmp_v3.auth.proto_len;
			memcpy(sess.securityAuthKey, 
			       srv->snmp_v3.auth.key,
			       srv->snmp_v3.auth.key_len);
			sess.securityAuthKeyLen =
				srv->snmp_v3.auth.key_len;
			
			sess.securityPrivProto =
				srv->snmp_v3.priv.proto;
			sess.securityPrivProtoLen =
				srv->snmp_v3.priv.proto_len;
			memcpy(sess.securityPrivKey,
			       srv->snmp_v3.priv.key,
			       srv->snmp_v3.priv.key_len);
			sess.securityPrivKeyLen =
				srv->snmp_v3.priv.key_len;
			if (srv->snmp_v3.user) {
				sess.securityName = srv->snmp_v3.user;
				sess.securityNameLen =
					strlen(sess.securityName);
			}
			if (srv->snmp_v3.context_name) {
				sess.contextName =
					srv->snmp_v3.context_name;
				sess.contextNameLen =
					strlen(srv->snmp_v3.context_name);
			}
			if (srv->snmp_v3.security_engine_id) {
				sess.securityEngineID =
					srv->snmp_v3.security_engine_id;
				sess.securityEngineIDLen =
					srv->snmp_v3.security_engine_id_len;
			}
			if (srv->snmp_v3.context_engine_id) {
				sess.contextEngineID =
					srv->snmp_v3.context_engine_id;
				sess.contextEngineIDLen =
					srv->snmp_v3.context_engine_id_len;
			}
			sess.securityLevel = srv->snmp_v3.security_level;
			sess.engineBoots = srv->snmp_v3.engine_boots;
			sess.engineTime = srv->snmp_v3.engine_time;
		}
		
		sess.callback = asynch_response;
		sess.callback_magic = srv;
		if (!(srv->sess = snmp_open(&sess))) {
			logmsg(LOG_ERR, "server %s: snmp_open failed: %s",
			       srvid(srv), snmp_api_errstring(snmp_errno));
			continue;
		}

		varinst_unset(srv->varinst);
		srv->flags &= ~SLB_SRV_COMPUTED;
		
		req = snmp_pdu_create(SNMP_MSG_GET);
		grecs_symtab_enumerate(srv->varinst, add_snmp_var, req);
		grecs_symtab_enumerate(srv->assertions, add_snmp_assertion,
				       req);
		
		if (snmp_send(srv->sess, req)) {
			active_count++;
		} else {
			logmsg(LOG_ERR, "server %s: snmp_send failed: %s",
			       srvid(srv), snmp_api_errstring(snmp_errno));
			snmp_free_pdu(req);
			snmp_close(srv->sess);
			srv->sess = NULL;
		}
	}
}

static void
recv_loop()
{
	debug(SLB_DEBCAT_SNMP, 1, ("waiting for replies"));
	while (active_count) {
		int fds = 0, block = 1;
		fd_set fdset;
		struct timeval timeout;

		FD_ZERO(&fdset);
		snmp_select_info(&fds, &fdset, &timeout, &block);
		fds = select(fds, &fdset, NULL, NULL, block ? NULL : &timeout);
		if (fds)
			snmp_read(&fdset);
		else
			snmp_timeout();
	}
}

static void
sess_cleanup()
{
	struct slb_server *srv;
	
	for (srv = srv_head; srv; srv = srv->next) {
		if (srv->sess) {
			snmp_close(srv->sess);
			srv->sess = NULL;
		}
	}
}

static int
srvcmp(const void *a, const void *b)
{
	struct slb_server const *sa = *(struct slb_server const **)a;
	struct slb_server const *sb = *(struct slb_server const **)b;

	if (sa->weight < sb->weight)
		return -1;
	if (sa->weight > sb->weight)
		return 1;
	return 0;
}

void
sort()
{
	struct slb_server *srv;
	size_t i;

	if (!srvtab)
		srvtab = grecs_calloc(srv_count, sizeof(*srv));
	
	for (srv = srv_head, i = 0; srv; srv = srv->next)
		if (srv->flags & SLB_SRV_COMPUTED)
			srvtab[i++] = srv;
	srvtab_count = i;
	
	qsort(srvtab, srvtab_count, sizeof(srvtab[0]), srvcmp);

	debug(SLB_DEBCAT_SNMP, 1,
	      ("load table computed (%d entries)", srvtab_count));
	
	if (debug_level[SLB_DEBCAT_SNMP] >= 4) {
		debug_printf("servers by load:");
		for (i = 0; i < srvtab_count; i++)
			debug_printf("%s %g", srvid(srvtab[i]),
				     srvtab[i]->weight);
	}
}
	

void
snmploop()
{
	active_count = 0;
	slb_loop_serial++;
	send_requests();
	slb_loop_ts = time(NULL);
	recv_loop();
	sess_cleanup();
	sort();
}
